from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_action_item.models.action_item import ActionItem
from edc_locator.models import SubjectLocator
from edc_protocol import Protocol
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from model_bakery import baker

from edc_consent.exceptions import NotConsentedError
from edc_consent.requires_consent import RequiresConsent
from edc_consent.site_consents import SiteConsentError, site_consents

from ..consent_test_utils import consent_object_factory
from ..models import CrfOne, SubjectVisit
from ..visit_schedules import visit_schedule


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=False,
)
class TestRequiresConsent(TestCase):
    def setUp(self):
        super().setUp()
        site_consents.registry = {}
        self.subject_identifier = "12345"
        self.study_open_datetime = Protocol().study_open_datetime
        self.study_close_datetime = Protocol().study_close_datetime

    def test_(self):
        self.assertRaises(SiteConsentError, RequiresConsent)

    def test_consent_out_of_period(self):
        consent_object_factory(start=self.study_open_datetime, end=self.study_close_datetime)
        self.assertRaises(
            SiteConsentError,
            baker.make_recipe,
            "edc_consent.subjectconsent",
            subject_identifier=self.subject_identifier,
            consent_datetime=self.study_close_datetime + relativedelta(days=1),
        )

    def test_not_consented(self):
        consent_object_factory(start=self.study_open_datetime, end=self.study_close_datetime)
        self.assertRaises(
            NotConsentedError,
            RequiresConsent,
            model="edc_consent.testmodel",
            subject_identifier=self.subject_identifier,
            consent_model="edc_consent.subjectconsent",
            report_datetime=self.study_open_datetime,
        )

    def test_consented(self):
        consent_object_factory(start=self.study_open_datetime, end=self.study_close_datetime)
        baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=self.subject_identifier,
            consent_datetime=self.study_open_datetime + relativedelta(months=1),
        )
        try:
            RequiresConsent(
                model="edc_consent.testmodel",
                subject_identifier=self.subject_identifier,
                consent_model="edc_consent.subjectconsent",
                report_datetime=self.study_open_datetime,
            )
        except NotConsentedError:
            self.fail("NotConsentedError unexpectedly raised")

    def test_requires_consent(self):
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)
        consent_object_factory(start=self.study_open_datetime, end=self.study_close_datetime)
        consent_obj = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=self.subject_identifier,
            consent_datetime=self.study_open_datetime + relativedelta(months=1),
        )
        subject_visit = SubjectVisit.objects.create(subject_identifier=self.subject_identifier)
        self.assertRaises(
            SiteConsentError,
            CrfOne.objects.create,
            subject_visit=subject_visit,
            subject_identifier="12345",
            report_datetime=self.study_close_datetime + relativedelta(months=1),
        )
        subject_visit = SubjectVisit.objects.create(subject_identifier=self.subject_identifier)
        try:
            CrfOne.objects.create(
                subject_visit=subject_visit,
                subject_identifier="12345",
                report_datetime=self.study_open_datetime + relativedelta(months=1),
            )
        except (SiteConsentError, NotConsentedError) as e:
            self.fail(f"Exception unexpectedly raised. Got {e}")
        consent_obj.delete()
        subject_visit = SubjectVisit.objects.create(subject_identifier=self.subject_identifier)
        self.assertRaises(
            NotConsentedError,
            CrfOne.objects.create,
            subject_visit=subject_visit,
            subject_identifier="12345",
            report_datetime=self.study_open_datetime + relativedelta(months=1),
        )
        self.assertRaises(
            NotConsentedError,
            SubjectLocator.objects.create,
            subject_identifier="12345",
            report_datetime=self.study_open_datetime - relativedelta(months=1),
        )

        # delete singleton action item
        # created for the subject locator
        ActionItem.objects.all().delete()

        try:
            SubjectLocator.objects.create(
                subject_identifier="12345",
                report_datetime=self.study_open_datetime + relativedelta(months=1),
            )
        except NotConsentedError as e:
            self.fail(f"NotConsentedError unexpectedly raised. Got {e}")



from datetime import datetime
from typing import Optional

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from edc_utils import convert_php_dateformat

from .exceptions import NotConsentedError
from .site_consents import SiteConsentError, site_consents


class RequiresConsent:
    def __init__(
        self,
        model: Optional[str] = None,
        subject_identifier: Optional[str] = None,
        report_datetime: Optional[datetime] = None,
        consent_model: Optional[str] = None,
        consent_group: Optional[str] = None,
    ):
        self.version = None
        self.model = model
        self.subject_identifier = subject_identifier
        self.consent_model = consent_model
        self.report_datetime = report_datetime
        self.consent_object = site_consents.get_consent_for_period(
            model=consent_model,
            consent_group=consent_group,
            report_datetime=report_datetime,
        )
        self.consent_model_cls = self.consent_object.model_cls
        self.version = self.consent_object.version
        if not self.subject_identifier:
            raise SiteConsentError(
                f"Cannot lookup {self.consent_model} instance for subject. "
                f"Got 'subject_identifier' is None."
            )
        self.consented_or_raise()

    def consented_or_raise(self):
        try:
            self.consent_model_cls.objects.get(
                subject_identifier=self.subject_identifier, version=self.version
            )
        except ObjectDoesNotExist:
            formatted_report_datetime = self.report_datetime.strftime(
                convert_php_dateformat(settings.SHORT_DATE_FORMAT)
            )
            raise NotConsentedError(
                f"Consent is required. Cannot find '{self.consent_model} "
                f"version {self.version}' when saving model '{self.model}' for "
                f"subject '{self.subject_identifier}' with date "
                f"'{formatted_report_datetime}'. "
                f"See also `all_post_consent_models` in the visit schedule."
            )



import sys

from django.apps import AppConfig as DjangoAppConfig

from .constants import DEFAULT_CONSENT_GROUP


class AppConfig(DjangoAppConfig):
    name = "edc_consent"
    verbose_name = "Edc Consent"
    default_consent_group = DEFAULT_CONSENT_GROUP
    include_in_administration_section = True

    def ready(self):
        from .site_consents import site_consents

        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        site_consents.autodiscover()
        for consent in site_consents.consents:
            start = consent.start.strftime("%Y-%m-%d %Z")
            end = consent.end.strftime("%Y-%m-%d %Z")
            sys.stdout.write(f" * {consent} covering {start} to {end}\n")
        sys.stdout.write(f" Done loading {self.verbose_name}.\n")



from uuid import uuid4

from django.db import models
from django.db.models import options
from django_crypto_fields.fields import EncryptedTextField
from edc_constants.constants import OPEN
from edc_data_manager.get_data_queries import get_data_queries
from edc_model.validators import datetime_not_future
from edc_protocol.validators import datetime_not_before_study_start
from edc_sites.models import CurrentSiteManager
from edc_utils import age, formatted_age

from ..consent_helper import ConsentHelper
from ..constants import DEFAULT_CONSENT_GROUP
from ..field_mixins import VerificationFieldsMixin
from ..managers import ConsentManager, ObjectConsentManager

if "consent_group" not in options.DEFAULT_NAMES:
    options.DEFAULT_NAMES = options.DEFAULT_NAMES + ("consent_group",)


class ConsentModelMixin(VerificationFieldsMixin, models.Model):
    """Mixin for a Consent model class such as SubjectConsent.

    Declare with edc_identifier's NonUniqueSubjectIdentifierModelMixin
    """

    consent_helper_cls = ConsentHelper

    consent_datetime = models.DateTimeField(
        verbose_name="Consent date and time",
        validators=[datetime_not_before_study_start, datetime_not_future],
    )

    report_datetime = models.DateTimeField(null=True, editable=False)

    version = models.CharField(
        verbose_name="Consent version",
        max_length=10,
        help_text="See 'Consent Type' for consent versions by period.",
        editable=False,
    )

    updates_versions = models.BooleanField(default=False)

    sid = models.CharField(
        verbose_name="SID",
        max_length=15,
        null=True,
        blank=True,
        editable=False,
        help_text="Used for randomization against a prepared rando-list.",
    )

    comment = EncryptedTextField(verbose_name="Comment", max_length=250, blank=True, null=True)

    dm_comment = models.CharField(
        verbose_name="Data Management comment",
        max_length=150,
        null=True,
        editable=False,
        help_text="see also edc.data manager.",
    )

    consent_identifier = models.UUIDField(
        default=uuid4,
        editable=False,
        help_text="A unique identifier for this consent instance",
    )

    objects = ObjectConsentManager()

    consent = ConsentManager()

    on_site = CurrentSiteManager()

    def __str__(self):
        return f"{self.get_subject_identifier()} v{self.version}"

    def natural_key(self):
        return (self.get_subject_identifier_as_pk(),)  # noqa

    def save(self, *args, **kwargs):
        self.report_datetime = self.consent_datetime
        helper = self.get_consent_helper()
        self.version = helper.version
        self.updates_versions = True if helper.updates_versions else False
        super().save(*args, **kwargs)

    def get_consent_helper(self):
        """Returns a consent helper instance"""
        return self.consent_helper_cls(
            model_cls=self.__class__, update_previous=True, **self.__dict__
        )

    def get_subject_identifier(self):
        """Returns the subject_identifier"""
        try:
            return self.subject_identifier  # noqa
        except AttributeError as e:
            if "subject_identifier" in str(e):
                raise NotImplementedError(f"Missing model mixin. Got `{str(e)}`.")
            raise

    def get_subject_identifier_as_pk(self):
        """Returns the subject_identifier_as_pk"""
        try:
            return self.subject_identifier_as_pk  # noqa
        except AttributeError as e:
            if "subject_identifier_as_pk" in str(e):
                raise NotImplementedError(f"Missing model mixin. Got `{str(e)}`.")
            raise

    def get_dob(self):
        """Returns the date of birth"""
        try:
            return self.dob  # noqa
        except AttributeError as e:
            if "dob" in str(e):
                raise NotImplementedError(f"Missing model mixin. Got `{str(e)}`.")
            raise

    @property
    def age_at_consent(self):
        """Returns a relativedelta."""
        try:
            return age(self.get_dob(), self.consent_datetime)
        except AttributeError as e:
            if "dob" in str(e):
                raise NotImplementedError(f"Missing model mixin. Got `{str(e)}`.")
            raise

    @property
    def formatted_age_at_consent(self):
        """Returns a string representation."""
        return formatted_age(self.get_dob(), self.consent_datetime)

    @property
    def open_data_queries(self):
        return get_data_queries(
            subject_identifier=self.subject_identifier,
            model=self._meta.label_lower,
            status=OPEN,
        )

    class Meta:
        abstract = True
        verbose_name = "Subject Consent"
        verbose_name_plural = "Subject Consents"
        consent_group = DEFAULT_CONSENT_GROUP
        get_latest_by = "consent_datetime"
        unique_together = (
            ("first_name", "dob", "initials", "version"),
            ("subject_identifier", "version"),
        )
        ordering = ("created",)

        indexes = [
            models.Index(
                fields=[
                    "subject_identifier",
                    "first_name",
                    "dob",
                    "initials",
                    "version",
                ]
            )
        ]



from datetime import datetime, timedelta

from dateutil.relativedelta import relativedelta
from django.test import TestCase, override_settings
from edc_protocol import Protocol
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from model_bakery import baker

from edc_consent.consent import NaiveDatetimeError
from edc_consent.consent_object_validator import (
    ConsentPeriodError,
    ConsentPeriodOverlapError,
    ConsentVersionSequenceError,
)
from edc_consent.exceptions import NotConsentedError
from edc_consent.site_consents import SiteConsentError, site_consents

from ..consent_test_utils import consent_object_factory
from ..models import CrfOne, SubjectVisit
from ..visit_schedules import visit_schedule


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=False,
)
class TestConsent(TestCase):
    def setUp(self):
        site_consents.registry = {}
        self.study_open_datetime = Protocol().study_open_datetime
        self.study_close_datetime = Protocol().study_close_datetime
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)
        self.subject_identifier = "12345"
        super().setUp()

    def test_raises_error_if_no_consent(self):
        """Asserts SubjectConsent cannot create a new instance if
        no consents are defined.

        Note: site_consents.reset_registry called in setUp.
        """
        subject_identifier = self.subject_identifier
        self.assertRaises(
            SiteConsentError,
            baker.make_recipe,
            "edc_consent.subjectconsent",
            subject_identifier=subject_identifier,
            consent_datetime=self.study_open_datetime,
        )

    def test_raises_error_if_no_consent2(self):
        """Asserts a model using the RequiresConsentMixin cannot create
        a new instance if subject not consented.
        """
        consent_object_factory(start=self.study_open_datetime, end=self.study_close_datetime)
        RegisteredSubject.objects.create(subject_identifier=self.subject_identifier)
        subject_visit = SubjectVisit.objects.create(subject_identifier=self.subject_identifier)
        self.assertRaises(
            NotConsentedError,
            CrfOne.objects.create,
            subject_visit=subject_visit,
            subject_identifier=self.subject_identifier,
            report_datetime=self.study_open_datetime,
        )

    def test_allows_create_if_consent(self):
        """Asserts can create a consent model instance if a valid
        consent.
        """
        consent_object_factory(start=self.study_open_datetime, end=self.study_close_datetime)
        baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=self.subject_identifier,
            consent_datetime=self.study_open_datetime,
            dob=self.study_open_datetime - relativedelta(years=25),
        )
        subject_visit = SubjectVisit.objects.create(subject_identifier=self.subject_identifier)
        try:
            CrfOne.objects.create(
                subject_visit=subject_visit,
                subject_identifier=self.subject_identifier,
                report_datetime=self.study_open_datetime,
            )
        except NotConsentedError:
            self.fail("NotConsentedError unexpectedly raised")

    def test_cannot_create_consent_without_consent_by_datetime(self):
        consent_object_factory(
            start=self.study_open_datetime + relativedelta(days=5),
            end=self.study_close_datetime,
            version="1",
        )
        self.assertRaises(
            SiteConsentError,
            baker.make_recipe,
            "edc_consent.subjectconsent",
            dob=self.study_open_datetime - relativedelta(years=25),
            consent_datetime=self.study_open_datetime,
        )

    def test_consent_gets_version(self):
        consent_object_factory(
            start=self.study_open_datetime, end=self.study_close_datetime, version="1.0"
        )
        consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            consent_datetime=self.study_open_datetime,
            dob=self.study_open_datetime - relativedelta(years=25),
        )
        self.assertEqual(consent.version, "1.0")

    def test_model_gets_version(self):
        consent_object_factory(
            start=self.study_open_datetime, end=self.study_close_datetime, version="1.0"
        )
        baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=self.subject_identifier,
            consent_datetime=self.study_open_datetime,
            dob=self.study_open_datetime - relativedelta(years=25),
        )
        subject_visit = SubjectVisit.objects.create(subject_identifier=self.subject_identifier)
        crf_one = CrfOne.objects.create(
            subject_visit=subject_visit,
            subject_identifier=self.subject_identifier,
            report_datetime=self.study_open_datetime,
        )
        self.assertEqual(crf_one.consent_version, "1.0")

    def test_model_consent_version_no_change(self):
        consent_object_factory(
            start=self.study_open_datetime, end=self.study_close_datetime, version="1.2"
        )
        baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=self.subject_identifier,
            consent_datetime=self.study_open_datetime,
            dob=self.study_open_datetime - relativedelta(years=25),
        )
        subject_visit = SubjectVisit.objects.create(subject_identifier=self.subject_identifier)
        crf_one = CrfOne.objects.create(
            subject_visit=subject_visit,
            subject_identifier=self.subject_identifier,
            report_datetime=self.study_open_datetime,
        )
        self.assertEqual(crf_one.consent_version, "1.2")
        crf_one.save()
        self.assertEqual(crf_one.consent_version, "1.2")

    def test_model_consent_version_changes_with_report_datetime(self):
        consent_object_factory(
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        consent_object_factory(
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="1.1",
        )
        consent_datetime = self.study_open_datetime + timedelta(days=10)
        subject_consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=self.subject_identifier,
            consent_datetime=consent_datetime,
            dob=self.study_open_datetime - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.version, "1.0")
        self.assertEqual(subject_consent.subject_identifier, self.subject_identifier)
        self.assertEqual(subject_consent.consent_datetime, consent_datetime)
        subject_visit = SubjectVisit.objects.create(subject_identifier=self.subject_identifier)
        crf_one = CrfOne.objects.create(
            subject_visit=subject_visit,
            subject_identifier=self.subject_identifier,
            report_datetime=consent_datetime,
        )
        self.assertEqual(crf_one.consent_version, "1.0")
        consent_datetime = self.study_open_datetime + timedelta(days=60)
        subject_consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=self.subject_identifier,
            consent_datetime=consent_datetime,
            dob=self.study_open_datetime - relativedelta(years=25),
        )
        crf_one.report_datetime = consent_datetime
        crf_one.save()
        self.assertEqual(crf_one.consent_version, "1.1")

    def test_consent_update_needs_previous_version(self):
        """Asserts that a consent type updates a previous consent."""
        consent_object_factory(
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        # specify updates version that does not exist, raises
        self.assertRaises(
            ConsentVersionSequenceError,
            consent_object_factory,
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="1.1",
            updates_versions="1.2",
        )
        # specify updates version that exists, ok
        consent_object_factory(
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="1.1",
            updates_versions="1.0",
        )

    def test_consent_model_needs_previous_version(self):
        """Asserts that a consent updates a previous consent but cannot
        be entered without an existing instance for the previous
        version."""
        consent_object_factory(
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        consent_object_factory(
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="1.1",
            updates_versions="1.0",
        )
        self.assertRaises(
            ConsentVersionSequenceError,
            baker.make_recipe,
            "edc_consent.subjectconsent",
            dob=self.study_open_datetime - relativedelta(years=25),
            consent_datetime=self.study_open_datetime + timedelta(days=60),
        )

    def test_consent_needs_previous_version2(self):
        """Asserts that a consent model updates its previous consent."""
        consent_object_factory(
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        consent_object_factory(
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="1.1",
            updates_versions="1.0",
        )
        subject_consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            consent_datetime=self.study_open_datetime + timedelta(days=5),
            dob=self.study_open_datetime - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.version, "1.0")
        subject_consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            subject_identifier=subject_consent.subject_identifier,
            consent_datetime=self.study_open_datetime + timedelta(days=60),
            first_name=subject_consent.first_name,
            last_name=subject_consent.last_name,
            initials=subject_consent.initials,
            identity=subject_consent.identity,
            confirm_identity=subject_consent.identity,
            dob=subject_consent.dob,
        )
        self.assertEqual(subject_consent.version, "1.1")

    def test_consent_needs_previous_version3(self):
        """Asserts that a consent updates a previous consent raises
        if a version is skipped.
        """
        consent_object_factory(
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        consent_object_factory(
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="1.1",
            updates_versions="1.0",
        )
        consent_object_factory(
            start=self.study_open_datetime + timedelta(days=101),
            end=self.study_open_datetime + timedelta(days=150),
            version="1.2",
            updates_versions="1.1",
        )
        subject_consent = baker.make_recipe(
            "edc_consent.subjectconsent",
            consent_datetime=self.study_open_datetime,
            dob=self.study_open_datetime - relativedelta(years=25),
        )
        self.assertEqual(subject_consent.version, "1.0")
        # use a consent datetime within verion 1.2, skipping 1.1, raises
        self.assertRaises(
            ConsentVersionSequenceError,
            baker.make_recipe,
            "edc_consent.subjectconsent",
            consent_datetime=self.study_open_datetime + timedelta(days=125),
            subject_identifier=subject_consent.subject_identifier,
            first_name=subject_consent.first_name,
            last_name=subject_consent.last_name,
            initials=subject_consent.initials,
            identity=subject_consent.identity,
            confirm_identity=subject_consent.identity,
            dob=subject_consent.dob,
        )

    def test_consent_periods_cannot_overlap(self):
        consent_object_factory(
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        self.assertRaises(
            ConsentPeriodOverlapError,
            consent_object_factory,
            start=self.study_open_datetime + timedelta(days=25),
            end=self.study_open_datetime + timedelta(days=100),
            version="1.1",
            updates_versions="1.0",
        )

    def test_consent_periods_cannot_overlap2(self):
        consent_object_factory(
            model="edc_consent.subjectconsent",
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        self.assertRaises(
            ConsentPeriodOverlapError,
            consent_object_factory,
            model="edc_consent.subjectconsent",
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.1",
        )

    def test_consent_periods_can_overlap_if_different_model(self):
        consent_object_factory(
            model="edc_consent.subjectconsent",
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        try:
            consent_object_factory(
                model="edc_consent.subjectconsent2",
                start=self.study_open_datetime,
                end=self.study_open_datetime + timedelta(days=50),
                version="1.0",
            )
        except ConsentPeriodOverlapError:
            self.fail("ConsentPeriodOverlapError unexpectedly raised")

    def test_consent_before_open(self):
        """Asserts cannot register a consent with a start date
        before the study open date.
        """
        self.assertRaises(
            ConsentPeriodError,
            consent_object_factory,
            start=self.study_open_datetime - relativedelta(days=1),
            end=self.study_close_datetime + relativedelta(days=1),
            version="1.0",
        )

    def test_consent_may_update_more_than_one_version(self):
        consent_object_factory(
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        consent_object_factory(
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="2.0",
        )
        consent_object_factory(
            start=self.study_open_datetime + timedelta(days=101),
            end=self.study_open_datetime + timedelta(days=150),
            version="3.0",
            updates_versions="1.0, 2.0",
        )

    def test_consent_object_naive_datetime_start(self):
        """Asserts cannot register a consent with a start date
        before the study open date.
        """
        d = self.study_open_datetime
        dte = datetime(d.year, d.month, d.day, 0, 0, 0, 0)
        self.assertRaises(
            NaiveDatetimeError,
            consent_object_factory,
            start=dte,
            end=self.study_close_datetime + relativedelta(days=1),
            version="1.0",
        )

    def test_consent_object_naive_datetime_end(self):
        """Asserts cannot register a consent with a start date
        before the study open date.
        """
        d = self.study_close_datetime
        dte = datetime(d.year, d.month, d.day, 0, 0, 0, 0)
        self.assertRaises(
            NaiveDatetimeError,
            consent_object_factory,
            start=self.study_open_datetime,
            end=dte,
            version="1.0",
        )



from __future__ import annotations

from datetime import datetime

from edc_consent.consent import Consent
from edc_consent.site_consents import site_consents


def consent_object_factory(
    model: str | None = None,
    start: datetime = None,
    end: datetime = None,
    gender: list[str] | None = None,
    updates_versions: list = None,
    version: str | None = None,
    age_min: int | None = None,
    age_max: int | None = None,
    age_is_adult: int | None = None,
):
    options = dict(
        start=start,
        end=end,
        gender=gender or ["M", "F"],
        updates_versions=updates_versions or [],
        version=version or "1",
        age_min=age_min or 16,
        age_max=age_max or 64,
        age_is_adult=age_is_adult or 18,
    )
    model = model or "edc_consent.subjectconsent"
    consent = Consent(model, **options)
    site_consents.register(consent)
    return consent


def consent_factory(model=None, **kwargs):
    options = dict(
        start=kwargs.get("start"),
        end=kwargs.get("end"),
        gender=kwargs.get("gender", ["M", "F"]),
        updates_versions=kwargs.get("updates_versions", []),
        version=kwargs.get("version", "1"),
        age_min=kwargs.get("age_min", 16),
        age_max=kwargs.get("age_max", 64),
        age_is_adult=kwargs.get("age_is_adult", 18),
    )
    model = kwargs.get("model", model or "edc_consent.subjectconsent")
    consent = Consent(model, **options)
    site_consents.register(consent)
    return consent



from __future__ import annotations

import sys
from copy import deepcopy
from datetime import datetime
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.conf import settings
from django.core.management.color import color_style
from django.utils.module_loading import import_module, module_has_submodule
from edc_utils import convert_php_dateformat, floor_secs

from .consent_object_validator import ConsentObjectValidator
from .exceptions import ConsentObjectDoesNotExist

if TYPE_CHECKING:
    from .consent import Consent


class ConsentError(Exception):
    pass


class AlreadyRegistered(Exception):
    pass


class SiteConsentError(Exception):
    pass


class SiteConsents:
    validate_consent_object = ConsentObjectValidator

    def __init__(self):
        self.registry = {}
        self.loaded = False

    def register(self, consent=None):
        """Register consent `objects`, not models."""
        if consent.name in self.registry:
            raise AlreadyRegistered(f"Consent object already registered. Got {consent}.")
        self.validate_consent_object(consent=consent, consents=self.consents)
        self.registry.update({consent.name: consent})
        self.loaded = True

    @property
    def consents(self) -> list[Consent]:
        """Returns an ordered list of consent objects."""
        consents = list(self.registry.values())
        return sorted(consents, key=lambda x: x.name, reverse=False)

    def get_consents_by_model(self, model: str = None) -> list[Consent] | list[str]():
        """Returns a list of consents for the given
        consent model label_lower.
        """
        return [consent for consent in self.consents if consent.model == model]

    def get_consent_for_period(
        self,
        model: str = None,
        report_datetime: datetime = None,
        consent_group: str | None = None,
    ) -> Consent:
        """Returns a consent object with a date range that the
        given report_datetime falls within.
        """
        if not self.consents or not self.loaded:
            raise SiteConsentError(
                f"No consent objects have been registered with `site_consents`. "
                f"Got {self.consents}, loaded={self.loaded}."
            )
        app_config = django_apps.get_app_config("edc_consent")
        consent_group = consent_group or app_config.default_consent_group
        registered_consents = self.registry.values()
        registered_consents = [
            c for c in registered_consents if c.group == consent_group and c.model == model
        ]
        if not registered_consents:
            raise SiteConsentError(
                f"No matching registered consent object in site consents. "
                f"Got consent_model={model}, consent_group={consent_group}, "
                f"report_datetime={report_datetime}. ",
                f"Expected one of {self.consents}.",
            )
        registered_consents = [
            c for c in registered_consents if c.start <= report_datetime <= c.end
        ]
        if not registered_consents:
            raise SiteConsentError(
                f"Date does not fall within the period of any registered consent "
                f"object. Got {report_datetime}. Expected one of {self.consents}."
            )
        return registered_consents[0]

    def get_consent(
        self,
        consent_model: str = None,
        report_datetime: datetime = None,
        version=None,
        consent_group=None,
    ) -> Consent:
        """Return consent object, not model, valid for the datetime."""
        app_config = django_apps.get_app_config("edc_consent")
        consent_group = consent_group or app_config.default_consent_group
        registered_consents = self.registry.values()
        if consent_group:
            registered_consents = [c for c in registered_consents if c.group == consent_group]
        if not registered_consents:
            raise ConsentObjectDoesNotExist(
                f"No matching consent in site consents. " f"Got consent_group={consent_group}."
            )
        if version:
            registered_consents = [c for c in registered_consents if c.version == version]
        if not registered_consents:
            raise ConsentObjectDoesNotExist(
                f"No matching consent in site consents. "
                f"Got consent_group={consent_group}, version={version}."
            )
        if consent_model:
            registered_consents = [c for c in registered_consents if c.model == consent_model]
        if not registered_consents:
            raise ConsentObjectDoesNotExist(
                f"No matching consent in site consents. "
                f"Got consent_group={consent_group}, version={version}, "
                f"model={consent_model}."
            )
        registered_consents = [
            c
            for c in registered_consents
            if floor_secs(c.start) <= floor_secs(report_datetime) <= floor_secs(c.end)
        ]
        if not registered_consents:
            raise ConsentObjectDoesNotExist(
                f"No matching consent in site consents. "
                f"Got consent_group={consent_group}, version={version}, "
                f"model={consent_model}, report_datetime={report_datetime}."
            )
        elif len(registered_consents) > 1:
            consents = list(set([c.name for c in registered_consents]))
            formatted_report_datetime = report_datetime.strftime(
                convert_php_dateformat(settings.SHORT_DATE_FORMAT)
            )
            raise ConsentError(
                f"Multiple consents found, using consent model={consent_model}, "
                f"date={formatted_report_datetime}, "
                f"consent_group={consent_group}, version={version}. "
                f"Got {consents}"
            )
        return registered_consents[0]

    @staticmethod
    def autodiscover(module_name=None, verbose=True):
        """Autodiscovers consent classes in the consents.py file of
        any INSTALLED_APP.
        """
        before_import_registry = None
        module_name = module_name or "consents"
        writer = sys.stdout.write if verbose else lambda x: x
        style = color_style()
        writer(f" * checking for site {module_name} ...\n")
        for app in django_apps.app_configs:
            writer(f" * searching {app}           \r")
            try:
                mod = import_module(app)
                try:
                    before_import_registry = deepcopy(site_consents.registry)
                    import_module(f"{app}.{module_name}")
                    writer(f" * registered consents '{module_name}' from '{app}'\n")
                except ConsentError as e:
                    writer(f"   - loading {app}.consents ... ")
                    writer(style.ERROR(f"ERROR! {e}\n"))
                except ImportError as e:
                    site_consents.registry = before_import_registry
                    if module_has_submodule(mod, module_name):
                        raise SiteConsentError(str(e))
            except ImportError:
                pass


site_consents = SiteConsents()



from django.db import models
from edc_constants.choices import YES_NO


class SampleCollectionFieldsMixin(models.Model):
    may_store_genetic_samples = models.CharField(
        verbose_name=(
            "Does the participant agree that a portion of "
            "the blood sample that is taken be stored for genetic "
            "analysis?"
        ),
        max_length=25,
        choices=YES_NO,
    )

    may_store_samples = models.CharField(
        verbose_name=(
            "Does the participant agree to have samples " "stored after the study has ended"
        ),
        max_length=3,
        choices=YES_NO,
    )

    class Meta:
        abstract = True



from ...consent_helper import ConsentHelper
from ...site_consents import site_consents
from .clean_fields_modelform_mixin import CleanFieldsModelformMixin
from .custom_validation_mixin import CustomValidationMixin


class ConsentModelFormMixin(CleanFieldsModelformMixin, CustomValidationMixin):
    def clean(self) -> dict:
        cleaned_data = super().clean()
        self.validate_initials_with_full_name()
        self.validate_gender_of_consent()
        self.validate_is_literate_and_witness()
        self.validate_dob_relative_to_consent_datetime()
        self.validate_guardian_and_dob()

        options = dict(
            consent_model=self._meta.model._meta.label_lower,
            consent_group=self._meta.model._meta.consent_group,
            report_datetime=self.consent_datetime,
        )
        consent = site_consents.get_consent(**options)
        if consent.updates_versions:
            ConsentHelper(
                model_cls=self._meta.model,
                update_previous=False,
                **cleaned_data,
            )
        self.validate_identity_and_confirm_identity()
        self.validate_identity_with_unique_fields()
        self.validate_identity_plus_version_is_unique()
        return cleaned_data
